(proclaim '(optimize (speed 0) (safety 3) (debug 3)))

(defconstant +max-microcode-memory+ #x200)
(defconstant +max-memory+ #x10000)
(defconstant +call-instruction+ 0)
(defparameter *ram* (make-array +max-memory+
                                :element-type '(unsigned-byte 8) 
                                :initial-element 0))
(defparameter *rom* (make-array +max-microcode-memory+
                                :element-type '(unsigned-byte 8) 
                                :initial-element 0))

;;; current index into the *rom* array while building the microcode
(defparameter *rom-index* 256)

;;; maps program-names to program-indices in the upper *rom* array
(defparameter *program-indices* (make-hash-table :test 'equal))

;;; maps primitive-names to primitive-codes
(defparameter *primitives* (make-hash-table :test 'equal))

;;; maps primitive-codes to primitive-functions for emulator
(defparameter *primitive-functions* (make-hash-table :test 'equal))

;;; creates x"AD" from #xAD
(defun number-to-hex (number)
  (format nil "x\"~2,'0X\"" number))

(defun replace-minus-with-underscore (name)
  (loop for i from 0 below (length name) do
        (when (eql (aref name i) #\-)
          (setf (aref name i) #\_)))
  name)

(defun print-vhdl-constant (name code)
  (format t "constant mcode_~a : data_type := x\"~2,'0X\";~%"
          (replace-minus-with-underscore
           (string-downcase
            (symbol-name name)))
          code))

(defun microcode-to-function (microcode)
  (intern
   (string-upcase
    (concatenate 'string
                 "mc-"
                 (symbol-name microcode)))))

(loop for name in
      '(call
        load-pc
        load-address
        load-q
        load-accu
        load-x
        load-y
        store-pc
        store-address
        store-data
        store-accu
        store-x
        store-y
        inc
        dec
        add
        lshift-8
        or
        nop)
      for code = 0 then (1+ code) do
      (setf (gethash name *primitives*) code)
      (print-vhdl-constant name code)
      (setf (gethash code *primitive-functions*)
            (microcode-to-function name)))

(defun write-memory (byte)
  (setf (aref *rom* *rom-index*) byte)
  (incf *rom-index*))
  
(defun mark-return-instruction ()
  (let ((last-index (1- *rom-index*)))
    (setf (aref *rom* last-index)
          (logior (aref *rom* last-index) 128))))

(defun defprogram-impl (name-or-code commands)
  (let ((relative-index (- *rom-index* 256)))
    (if (numberp name-or-code)
        (setf (aref *rom* name-or-code) relative-index)
      (setf (gethash name-or-code *program-indices*) relative-index)))
  (loop for command in commands do
        (let ((program-index (gethash command *program-indices*))
              (primitive (gethash command *primitives*)))
          (cond (program-index (write-memory +call-instruction+)
                               (write-memory program-index))
                (primitive (write-memory primitive))
                (t (error "command not found: ~a" command)))))
  (mark-return-instruction))

(defmacro defprogram (name-or-code &body commands)
  (list 'defprogram-impl
        (list 'quote name-or-code)
        (list 'quote commands)))

(defprogram next-pc
  load-pc
  store-address
  load-pc
  inc
  store-pc
  load-q
  )
            
(defprogram next-pc-word
  next-pc
  next-pc
  lshift-8
  or
  )
            
(defprogram absolute
  next-pc-word
  store-address
  )

(defprogram absolute-x
  next-pc-word
  load-x
  add
  store-address
  )

(defprogram absolute-y
  next-pc-word
  load-y
  add
  store-address
  )

(defprogram immediate
  load-pc
  store-address
  load-pc
  inc
  store-pc
  )

(defprogram indirect
  next-pc-word
  store-address
  load-address
  inc
  store-address
  load-q
  load-q
  lshift-8
  or
  )

; LDA immediate
(defprogram #xa9
  immediate
  load-q
  store-accu
  )

; STA absolute
(defprogram #x8d
  absolute
  load-accu
  store-data
  )

; NOP
(defprogram #xea
  nop
  )

; store memory as MIF (memory initialization file)
(with-open-file (output "c:/tmp/test.mif"
                        :direction :output
                        :if-exists :supersede)
  (loop for i from 0 below +max-microcode-memory+ do
        (let ((byte (aref *rom* i)))
          (format output "~a : ~a;~%" i byte))))



;;;
;;; emulator
;;;

(defparameter *pc* #x200)
(defparameter *sp* #x1ff)
(defparameter *accu* 0)
(defparameter *x* 0)
(defparameter *y* 0)
(defparameter z-flag nil)
(defparameter n-flag nil)
(defparameter c-flag nil)
(defparameter v-flag nil)
(defparameter i-flag nil)
(defparameter d-flag nil)

(defparameter *address* 0)
(defparameter *data* 0)
(defparameter *q* 0)
(defparameter *wren* 0)

(defparameter *mcode-address* 0)
(defparameter *mcode-q* 0)

(defparameter *stack* nil)
(defparameter *call-stack* nil)

(defun next-mcode-byte ()
  (prog1 (aref *rom* *mcode-address*)
    (incf *mcode-address*)))

; implement primitives for emulator

(defun mc-call ()
  (let ((destination (+ #x100 (next-mcode-byte))))
    (push *mcode-address* *call-stack*)
    (setf *mcode-address* destination)))

(defun mc-load-pc ()
  (push *pc* *stack*))

(defun mc-load-address ()
  (push *address* *stack*))

(defun mc-load-q ()
  (push *q* *stack*))

(defun mc-load-accu ()
  (push *accu* *stack*))

(defun mc-load-x ()
  (push *x* *stack*))

(defun mc-load-y ()
  (push *y* *stack*))

(defun mc-store-pc ()
  (setf *pc* (pop *stack*)))

(defun mc-store-address ()
  (setf *address* (pop *stack*))
  (setf *q* (aref *ram* *address*)))

(defun mc-store-data ()
  (setf *data* (pop *stack*))
  (setf (aref *ram* *address*) *data*))

(defun mc-store-accu ()
  (setf *accu* (pop *stack*)))

(defun mc-store-x ()
  (setf *x* (pop *stack*)))

(defun mc-store-y ()
  (setf *y* (pop *stack*)))

(defun mc-inc ()
  (incf (first *stack*)))

(defun mc-dec ()
  (decf (first *stack*)))

(defun mc-add ()
  (push (+ (pop *stack*)
           (pop *stack*))
        *stack*))

(defun mc-lshift-8 ()
  (push (ash (pop *stack*) 8) *stack*))

(defun mc-or ()
  (push (logior (pop *stack*)
                (pop *stack*))
        *stack*))

(defun mc-nop ())

(defun dump-registers ()
  (format t "a: ~2,'0X, x: ~2,'0X, y: ~2,'0X~%" *accu* *x* *y*)
  (format t "pc: ~4,'0X, mcode-address: ~4,'0X,~%" *pc* *mcode-address*))

(defun execute-command ()
  (format t "current registers:")
  (terpri)
  (dump-registers)
  (terpri)
  (let ((command (aref *ram* *pc*)))
    (incf *pc*)
    (setf *mcode-address* (+ #x100 (aref *rom* command)))
    (loop for i = 0 then (1+ i) do
          (let ((microcode (next-mcode-byte))
                (return-instruction nil))
            (when (> microcode #x7f)
              (setf return-instruction t)
              (decf microcode #x80))
            (let ((mc-function (gethash microcode *primitive-functions*)))
              (format t "executing microcode: ~a~%"
                      (subseq (symbol-name mc-function) 3))
              (funcall mc-function)
              (dump-registers)
              (terpri))
            (when return-instruction
              (if *call-stack*
                  (setf *mcode-address* (pop *call-stack*))
                (loop-finish)))
            (when (> i 100)
              (format t "too many microinstructions?")
              (loop-finish))))))

(with-open-file (input "l:/cc65-2.11.0/test"
                       :direction :input
                       :element-type '(unsigned-byte 8))
  (let ((len (file-length input)))
    (loop for i from 0 below len do
          (setf (aref *ram* (+ i #x200)) (read-byte input)))))

(defun dump (address &optional (len 8))
  (loop for i from 0 below len do
        (let ((address (+ address i)))
          (when (= 0 (mod i 8))
            (terpri)
            (format t "~4,'0X: " address))
          (format t "~2,'0X " (aref *ram* address)))))
